// Copyright (C) 2022 The OKit Authors
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
// OR OTHER DEALINGS IN THE SOFTWARE.

package pb

import (
	"encoding/base64"
	"net/url"
	"strconv"
	"time"
)

//go:generate protoc --go_out=paths=source_relative:. -I=. obkit.proto

// AsDuration converts a protocol level Duration to a Golang time.Duration.
func (x *Duration) AsDuration() time.Duration {
	return time.Duration(x.Nanos)
}

// DurationPB converts a time.Duration to a protocol level Duration.
func DurationPB(d time.Duration) *Duration {
	return &Duration{
		Nanos: int64(d),
	}
}

// AsTime converts a protocol level Timestamp to a Golang time.Time.
func (x *Timestamp) AsTime() time.Time {
	return time.Unix(x.Seconds, int64(x.Nanos))
}

// TimestampPB converts a time.Time to a protocol level Timestamp.
func TimestampPB(t time.Time) *Timestamp {
	return &Timestamp{
		Seconds: t.Unix(),
		Nanos:   int32(t.Nanosecond()),
	}
}

// AsQueryString converts the associated tag to a URI query string.
func (x *Tag) AsQueryString() string {
	val := ""

	switch v := x.GetValue().(type) {
	case *Tag_String_:
		val = v.String_
	case *Tag_Int64:
		val = strconv.FormatInt(v.Int64, 10)
	case *Tag_Double:
		val = strconv.FormatFloat(v.Double, 'f', 9, 64)
	case *Tag_Bytes:
		val = base64.RawStdEncoding.EncodeToString(v.Bytes)
	case *Tag_Bool:
		val = strconv.FormatBool(v.Bool)
	case *Tag_Duration:
		val = v.Duration.AsDuration().String()
	case *Tag_Timestamp:
		val = v.Timestamp.AsTime().Format(time.RFC3339)
	}

	if val == "" {
		return ""
	}

	return url.QueryEscape(x.Key) + "=" + url.QueryEscape(val)
}
