// Copyright (C) 2022 The OKit Authors
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
// OR OTHER DEALINGS IN THE SOFTWARE.

package main

import (
	"flag"
	"io"
	"os"
	"text/template"
)

var format = `// Code generated by tools/gen-tags. DO NOT EDIT.
package {{ .Package }}

import (
	"time"

	"go.pitz.tech/okit/pb"
)

// Tag defines a generic data holder that allows values to be passed around during observation. Both raw values and
// pointers can be used, allowing data to be lazily read and pulled during emission.
type Tag struct {
	key   string
	value interface{}
}

// AsTagPB converts the Tag to a pb.Tag for transport and storage. If the value associated with the Tag is nil, then no
// tag is returned and the key is filtered from the list.
func (t Tag) AsTagPB() *pb.Tag {
	if t.value == nil {
		return nil
	}

	switch v := t.value.(type) {
	{{ range .Kinds -}}
	case {{ .ValueKind }}:
		return &pb.Tag{Key: t.key, Value: &pb.Tag_{{ .TagKind }}{ {{ .TagKind }}: {{ .Value }} }}
	{{ end -}}
	}

	return nil
}

{{ range .Kinds -}}
// {{ .Function }} returns a Tag whose value is read after submission.
func {{ .Function }}(key string, value {{ .ValueKind }}) Tag { return Tag{key, value} }

{{ end -}}
`

type Kind struct {
	Function  string
	ValueKind string
	TagKind   string
	Value     string
}

type Config struct {
	Package string
	Kinds   []Kind
}

func main() {
	out := flag.String("out", "", "")
	flag.Parse()

	var writer io.Writer = os.Stdout
	if out != nil && *out != "" {
		file, err := os.Create(*out)
		if err != nil {
			panic(err)
		}

		defer file.Close()
		writer = file
	}

	err := template.Must(template.New("format").Parse(format)).Execute(
		writer,
		Config{
			Package: "okit",
			Kinds: []Kind{
				{"String", "string", "String_", "v"},
				{"Stringp", "*string", "String_", "*v"},
				{"Int", "int", "Int64", "int64(v)"},
				{"Intp", "*int", "Int64", "int64(*v)"},
				{"Int8", "int8", "Int64", "int64(v)"},
				{"Int8p", "*int8", "Int64", "int64(*v)"},
				{"Int16", "int16", "Int64", "int64(v)"},
				{"Int16p", "*int16", "Int64", "int64(*v)"},
				{"Int32", "int32", "Int64", "int64(v)"},
				{"Int32p", "*int32", "Int64", "int64(*v)"},
				{"Int64", "int64", "Int64", "int64(v)"},
				{"Int64p", "*int64", "Int64", "int64(*v)"},
				{"Float32", "float32", "Double", "float64(v)"},
				{"Float32p", "*float32", "Double", "float64(*v)"},
				{"Float64", "float64", "Double", "float64(v)"},
				{"Float64p", "*float64", "Double", "float64(*v)"},
				{"Bytes", "[]byte", "Bytes", "v"},
				{"Bytesp", "*[]byte", "Bytes", "*v"},
				{"Bool", "bool", "Bool", "v"},
				{"Boolp", "*bool", "Bool", "*v"},
				{"Duration", "time.Duration", "Duration", "pb.DurationPB(v)"},
				{"Durationp", "*time.Duration", "Duration", "pb.DurationPB(*v)"},
				{"Timestamp", "time.Time", "Timestamp", "pb.TimestampPB(v)"},
				{"Timestampp", "*time.Time", "Timestamp", "pb.TimestampPB(*v)"},
				{"NamedErr", "error", "String_", "v.Error()"},
				{"NamedErrp", "*error", "String_", "(*v).Error()"},
			},
		},
	)

	if err != nil {
		panic(err)
	}
}
